;
;
.nolist
.include "tn261adef.inc" ; Define device ATtiny261A
.list
;// use ATtiny261A with stock fuse settings ($DF62)
;// running at 1MHz on internal oscillator

.macro blip ; toggle pin as debug marker etc.
sbi PORTA, PA2
cbi PORTA, PA2
.endmacro

.macro nop10 ; universal 10-clock delay
nop
nop
nop
nop
nop
nop
nop
nop
nop
nop
.endmacro

;// r0:r1 = NPN idle Ib for offset correction
;// r2:r3 = NPN Ib measurement
;// r4:r5 = PNP idle Ib for offset correction
;// r6:r7 = PNP Ib measurement
.def adcreg = R15 ; ADC
.def tmp = R16 ; Define multipurpose register
.def tmp2 = r17
.def tmp3 = r18
.def flag = r19 ; flag register


; Standard-Parameter-Satz der Einstellungen
.equ LcdClock = 1000000 ; Taktfrequenz Prozessor in Hz
.equ clock = 1000000 ; Taktfrequenz Prozessor in Hz
; LCD-Groesse:
.equ LcdLines = 2 ; Anzahl Zeilen (1, 2, 4)
.equ LcdCols = 16 ; Anzahl Zeichen pro Zeile (8..24)
; LCD-Ansteuerung
.equ LcdBits = 4 ; Busbreite (4 oder 8)
  ; Wenn 4-Bit-Ansteuerung:
.equ Lcd4High = 1 ; Busnibble (1=Oberes, 0=Unteres)
.equ LcdWait = 0 ; Ansteuerung (0 mit Busy, 1 mit Warteschleifen)
; LCD-Datenports
.equ pLcdDO = PORTA ; Daten-Ausgabe-Port
.equ pLcdDD = DDRA ; Daten-Richtungs-Port
; LCD-Kontrollports und -pins 1=RS 2=RW 3=E
.equ pLcdCEO = PORTB ; Control E Ausgabe-Port
.equ bLcdCEO = PORTB3 ; Controll E Ausgabe-Portpin
.equ pLcdCED = DDRB ; Control E Richtungs-Port
.equ bLcdCED = DDB3 ; Control E Richtungs-Portpin
.equ pLcdCRSO = PORTB ; Control RS Ausgabe-Port
.equ bLcdCRSO = PORTB1 ; Controll RS Ausgabe-Portpin
.equ pLcdCRSD = DDRB ; Control RS Richtungs-Port
.equ bLcdCRSD = DDB1 ; Control RS Richtungs-Portpin
; Wenn LcdWait = 0:
.equ pLcdDI = PINB ; Daten-Input-Port
.equ pLcdCRWO = PORTB ; Control RW Ausgabe-Port
.equ bLcdCRWO = PORTB2 ; Control RW Ausgabe-Portpin
.equ pLcdCRWD = DDRB ; Control RW Richtungs-Port
.equ bLcdCRWD = DDB2 ; Control RW Richtungs-Portpin
; Wenn Dezimalausgabe benoetigt wird:
.equ LcdDecimal = 1 ; Wenn definiert: einbinden
; Wenn Hexadezimalausgabe benoetigt wird:
.equ LcdHex = 1 ; Wenn definiert: einbinden
; Wenn nur Simulation im SRAM:
  ;.equ avr_sim = 1 ; 1=Simulieren, 0=Nicht simulieren
;
;
; **********************************
;           S R A M
; **********************************
;
.dseg
.org SRAM_START

.cseg
.org 000000
; **********************************
;  M A I N   P R O G R A M   I N I T
; **********************************
;
Main:
.ifdef SPH ; if SPH is defined
ldi tmp,High(RAMEND)
out SPH,tmp ; Init MSB stack pointer
.endif
ldi tmp,Low(RAMEND)
out SPL,tmp ; Init LSB stack pointer

ldi tmp, 0b11110100
out DDRA, tmp	; DDRx: 0=in 1=out
ldi tmp, 0b00111110
out DDRB, tmp	; DDRx: 0=in 1=out
; all outputs low
clr tmp
out PORTA, tmp	; PORTx: 1=in:pullup out:high
out PORTB, tmp	; PORTx: 1=in:pullup out:high
sbi PORTB, PB0  ; pullup on PB0

; init adc
;ldi tmp, (1<<REFS0)|(1<<REFS1) ;2.56V with bypass cap
;out ADMUX, tmp
;ldi tmp, (1<<REFS2) ;2.56V with bypass cap
;out ADCSRB, tmp
ldi tmp, (1<<REFS1) ;1.1V internal reference
out ADMUX, tmp
ldi tmp, (1<<ADC0D)|(1<<ADC1D)|(1<<AREFD) ;disable pin buffer
out DIDR0, tmp

;ldi tmp, (1<<ADEN)|(1<<ADSC)|(0<<ADPS1)|(1<<ADPS0) ;/2 = 500kHz
ldi tmp, (1<<ADEN)|(1<<ADSC)|(1<<ADPS1)|(0<<ADPS0) ;/4 = 250kHz
;ldi tmp, (1<<ADEN)|(1<<ADSC)|(1<<ADPS1)|(1<<ADPS0) ;/8 = 125kHz
mov adcreg, tmp
out ADCSRA, adcreg ;start first conversion

; init display
rcall LcdInit

rcall wait10
ldi ZH,High(2*InitText)
ldi ZL,Low(2*InitText)
rcall LcdText

; determine mode
clr flag
cbr flag, 0b01000000 ; normal mode, PB0 high through pullup, flag T bit zero
sbis PINB, PB0
sbr flag, 0b01000000 ; calibration mode, PB0 pulled low, flag T bit set

;branch on calibration mode
sbrc flag, 6
rjmp calmode


; ****************************************
; DEBUG
; ****************************************
;clr r20
;clr r21
;testloop:
;ldi ZL, 100
;testwait:
;dec ZL
;rcall wait10
;brne testwait
;ldi ZL, 1
;clr ZH
;; add 16bit: low byte first, then high byte with carry
;add r20, ZL
;adc r21, ZH
;clr r0
;clr r1
;mov r2, r20
;mov r3, r21
;clr r4
;clr r5
;mov r6, r20
;mov r7, r21
;rcall displaycal
;rjmp testloop


; ****************************************
; MAIN LOOP
; ****************************************

loop:
;poll if button is pressed
sbic PINB, PB6
rjmp loop
rcall wait10
sbic PINB, PB6
rjmp loop
rcall getbusy
released: sbis PINB, PB6
rjmp released
rcall wait10
rjmp loop


calmode:
;keep acquiring data, don't poll button
;keep CCS's active
sbi PORTB, PB4
sbi PORTB, PB5

rcall wait10
rcall wait10
rcall wait10
rcall wait10
rcall wait10
rcall getcal
rcall wait10
rcall wait10
rcall wait10
rcall wait10
rcall wait10

rcall wait10
rcall wait10
rcall wait10
rcall wait10
rcall wait10
rjmp calmode


rjmp loop


; ****************************************
; ACQUIRE SOME DATA
; ****************************************

getbusy:
;blip
; --------------------------------------------------------------------
; do NPN
cbi ADMUX, MUX0 ;select ADC input on PA0
out ADCSRA, adcreg ;start conversion
npn1adc:
sbic ADCSRA, ADSC
rjmp npn1adc
; fetch new reading. calibration run.
in r0, ADCL
in r1, ADCH
blip
; activate current source and leave it active for some microseconds
sbi PORTB, PB4
; NPN duration -------------------------------------------------------
nop10
nop10
nop10
;nop10
;nop10
; --------------------------------------------------------------------
blip
;; disable current source before ADC capture; the current through
;; the DUT will keep flowing long enough (~30µs)
cbi PORTB, PB4
out ADCSRA, adcreg ;start conversion
npn2adc:
sbic ADCSRA, ADSC
rjmp npn2adc
; fetch new reading. actual measurement data.
in r2, ADCL
in r3, ADCH
blip
; --------------------------------------------------------------------
; do PNP
sbi ADMUX, MUX0 ;select ADC input on PA1
out ADCSRA, adcreg ;start conversion
pnp1adc:
sbic ADCSRA, ADSC
rjmp pnp1adc
; fetch new reading. calibration run.
in r4, ADCL
in r5, ADCH
blip
; activate current source and leave it active for some microseconds
sbi PORTB, PB5
; PNP duration -------------------------------------------------------
nop10
nop10
nop10
;nop10
;nop10
; --------------------------------------------------------------------
blip
cbi PORTB, PB5
out ADCSRA, adcreg ;start conversion
pnp2adc:
sbic ADCSRA, ADSC
rjmp pnp2adc
; fetch new reading. actual measurement data.
in r6, ADCL
in r7, ADCH
blip
rcall displaydata
ret


; ****************************************
; ACQUIRE BASE CURRENT FOR CAL.MODE
; ****************************************
; run 16 averages
getcal:
; --------------------------------------------------------------------
; do NPN
cbi ADMUX, MUX0 ;select ADC input on PA0
ldi tmp, 16
clr r2
clr r3
loopnpncal:
out ADCSRA, adcreg ;start conversion
npn2adccal:
sbic ADCSRA, ADSC
rjmp npn2adccal
; fetch new reading. actual measurement data.
in r0, ADCL
in r1, ADCH
; add up
add r2, r0
adc r3, r1
dec tmp
brne loopnpncal
; divide by 16
lsr r3
ror r2
lsr r3
ror r2
lsr r3
ror r2
lsr r3
ror r2

; --------------------------------------------------------------------
; do PNP
sbi ADMUX, MUX0 ;select ADC input on PA0
ldi tmp, 16
clr r6
clr r7
looppnpcal:
out ADCSRA, adcreg ;start conversion
pnp2adccal:
sbic ADCSRA, ADSC
rjmp pnp2adccal
; fetch new reading. actual measurement data.
in r4, ADCL
in r5, ADCH
; add up
add r6, r4
adc r7, r5
dec tmp
brne looppnpcal
; divide by 16
lsr r7
ror r6
lsr r7
ror r6
lsr r7
ror r6
lsr r7
ror r6

rcall displaycal
ret


; ****************************************
; CALCULATE HFE
; ****************************************
; binary raw offset input in X
; binary raw data input in Y
; binary output in Z
hfecalc:
; remove offset X from Y
; subtract 16bit: low byte first, then high byte with carry
sub YL, XL
sbc YH, XH

; sanity check: discard > 1024 < 8 (16?)
movw XH:XL, YH:YL
sbiw XH:XL, 8
brlo discard
cpi YH, 4 ; check if high-byte is >= 4 (=1024)
brsh discard

; divide 152520 by Y
ldi tmp, 1
clr tmp2
clr ZL
clr ZH
ldi tmp3, 2
ldi XH, $53
ldi XL, $C8
divloop:
; subtract 24bit: low byte first, then high bytes with carry
sub XL, YL
sbc XH, YH
sbc tmp3, tmp2
brlo getfrac
; add 16bit: low byte first, then high byte with carry
add ZL, tmp
adc ZH, tmp2
rjmp divloop
getfrac:
;check remainder for more precise result
;add one back
add XL, YL
adc XH, YH

ret

discard:
; invalid data; return all zero
clr ZL
clr ZH
;ldi ZH, 3
;ldi ZL, 255
ret

; ****************************************
; CALCULATE BASE CURRENT
; ****************************************
; binary raw data input in Y
; binary output in Z
;
; ADC * 275 / 4096 = mA
; 275 = Vref 1100 / 4
;
; ADC * 1375 / 2048 = mA * 10
; 137.5 = Vref 1100 / 8
;
; use 16Ohm base resistor.
; use 24bit math (3 registers)

ibcalc:
; sanity check: discard > 1024 < 8 (16?)
movw XH:XL, YH:YL
sbiw XH:XL, 8
brlo discardib
cpi YH, 4 ; check if high-byte is >= 4 (=1024)
brsh discardib

; multiply Y by 1375
; MOVW only on pairs :(
; *256 by leaving out lowest register
mov tmp3, YH
mov ZH, YL
; *512
lsl ZH
rol tmp3
; *1024
lsl ZH
rol tmp3
; add 256 for *1280
add ZH, YL
adc tmp3, YH
; add 95 for *1375
ldi tmp, 95
clr tmp2
ibloop:
; add 24bit: low byte first, then high byte with carry
;add ZH, YL
;adc tmp3, YH
add ZL, YL
adc ZH, YH
adc tmp3, tmp2
dec tmp
;breq ibdone
;rjmp ibloop
;ibdone:
brne ibloop
; divide Z by 256
mov ZL, ZH
mov ZH, tmp3
; and another 3 times for 2048
lsr ZH
ror ZL
lsr ZH
ror ZL
lsr ZH
ror ZL
ret

discardib:
; invalid data; return all zero
clr ZL
clr ZH
ret

; ****************************************
; DISPLAY DATA
; ****************************************

displaydata:
; calculate and display hfe values
clr ZH
clr ZL
rcall LcdPos
ldi ZH,High(2*DataText)
ldi ZL,Low(2*DataText)
rcall LcdText

ldi ZH, 0
ldi ZL, 10
rcall LcdPos
mov XL, r0
mov XH, r1
mov YL, r2
mov YH, r3
rcall hfecalc
rcall LcdDec5

ldi ZH, 1
ldi ZL, 10
rcall LcdPos
mov XL, r4
mov XH, r5
mov YL, r6
mov YH, r7
rcall hfecalc
rcall LcdDec5
ret

displaycal:
; --------------------------------------------------------------------
; calculate and display base current values
clr ZH
clr ZL
rcall LcdPos
ldi ZH,High(2*CalText)
ldi ZL,Low(2*CalText)
rcall LcdText

ldi ZH, 0
ldi ZL, 10
rcall LcdPos
mov XL, r0
mov XH, r1
mov YL, r2
mov YH, r3
rcall ibcalc
rcall LcdDec5

ldi ZH, 1
ldi ZL, 10
rcall LcdPos
mov XL, r4
mov XH, r5
mov YL, r6
mov YH, r7
rcall ibcalc
rcall LcdDec5
ret

displayhex:
; --------------------------------------------------------------------
; display raw hex values
clr ZH
clr ZL
rcall LcdPos
ldi ZH,High(2*IdleText)
ldi ZL,Low(2*IdleText)
rcall LcdText

ldi ZH, 0
ldi ZL, 6
rcall LcdPos
mov ZL, r0
mov ZH, r1
rcall LcdHex4
ldi ZH, 0
ldi ZL, 12
rcall LcdPos
mov ZL, r2
mov ZH, r3
rcall LcdHex4

ldi ZH, 1
ldi ZL, 6
rcall LcdPos
mov ZL, r4
mov ZH, r5
rcall LcdHex4
ldi ZH, 1
ldi ZL, 12
rcall LcdPos
mov ZL, r6
mov ZH, r7
rcall LcdHex4
ret


; ****************************************
; WAIT ROUTINES (5ms, 10ms)
; ****************************************

wait5:
push r17
push r16
; 4ms 989us at 1 MHz
ldi  r16, 7
ldi  r17, 122
wait5l: dec  r17
brne wait5l
dec  r16
brne wait5l
pop r16
pop r17
ret

wait10:
push r17
push r16
; 9ms 987us at 1 MHz
ldi  r16, 13
ldi  r17, 248
wait10l: dec  r17
brne wait10l
dec  r16
brne wait10l
pop r16
pop r17
ret
;
;
.include "lcd.inc"

InitText:
.db "Press Button to ",0x0D,0xFF
.db "acquire reading ",0xFE,0xFF

IdleText:
.db "NPN             ",0x0D,0xFF
.db "PNP             ",0xFE,0xFF

DataText:
.db "NPN hFE =       ",0x0D,0xFF
.db "PNP hFE =       ",0xFE,0xFF

CalText:
.db "NPN Ib  =       ",0x0D,0xFF
.db "PNP Ib  =       ",0xFE,0xFF


